from qlat_utils import *
import qlat.c as c

from qlat.geometry import *
from qlat.utils_io import *

class PointSelection:

    # self.geo
    # self.cdata

    def __init__(self, coordinate_list = None, geo = None):
        if None == coordinate_list:
            self.cdata = c.mk_psel()
        else:
            self.cdata = c.mk_psel(coordinate_list)
        self.geo = geo

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_psel(self)

    def __imatmul__(self, v1):
        assert isinstance(v1, PointSelection)
        c.set_psel(self, v1)
        self.geo = v1.geo
        return self

    def copy(self):
        x = PointSelection()
        x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def set_rand(self, rs, total_site, n_points):
        c.set_rand_psel(self, rs, total_site, n_points)
        self.geo = Geometry(total_site)

    def save(self, path):
        c.save_psel(self, path)

    def load(self, path, geo = None):
        c.load_psel(self, path)
        self.geo = geo

    def n_points(self):
        return c.get_n_points_psel(self)

    def to_list(self):
        return c.mk_list_psel(self)

    def from_list(self, coordinate_list, geo = None):
        c.set_list_psel(self, coordinate_list)
        self.geo = geo
        return self

    def coordinate_from_idx(self, idx):
        return c.get_coordinate_from_idx_psel(self, idx)

cache_point_selection = mk_cache("point_selection")

def get_psel_tslice(total_site, *, t_dir = 3):
    # if t_dir = 3, then [ [0,0,0,0,], [0,0,0,1,], ..., [0,0,0,total_site[3]-1],]
    # if t_dir = 2, then [ [0,0,0,0,], [0,0,1,0,], ..., [0,0,total_site[2]-1],0,]
    # need total_site to set the psel.geo property
    assert 0 <= t_dir and t_dir < 4
    assert isinstance(total_site, list)
    param_tuple = (tuple(total_site), t_dir,)
    if param_tuple not in cache_point_selection:
        psel = PointSelection()
        c.set_tslice_psel(psel, total_site[t_dir], t_dir)
        psel.geo = Geometry(total_site)
        cache_point_selection[param_tuple] = psel
    return cache_point_selection[param_tuple]

class FieldSelection:

    # self.cdata

    def __init__(self, total_site = None, n_per_tslice = -1, rs = None, psel = None):
        self.cdata = c.mk_fsel()
        if total_site is not None:
            assert isinstance(rs, RngState)
            assert isinstance(n_per_tslice, int)
            c.set_rand_fsel(self, rs, total_site, n_per_tslice)
            if psel is not None:
                c.add_psel_fsel(self, psel)
            self.update()
            self.update(n_per_tslice)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_fsel(self)

    def __imatmul__(self, v1):
        assert isinstance(v1, FieldSelection)
        c.set_fsel(self, v1)
        return self

    def copy(self):
        x = FieldSelection()
        x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def set_uniform(self, total_site, val = 0):
        # default (val = 0) select every sites
        # val = -1 deselection everything
        c.set_uniform_fsel(self, total_site, val)

    def set_rand(self, rs, total_site, n_per_tslice):
        assert isinstance(rs, RngState)
        assert isinstance(n_per_tslice, int)
        c.set_rand_fsel(self, rs, total_site, n_per_tslice)
        self.update()
        self.update(n_per_tslice)

    def add_psel(self, psel, rank_psel = 1024 * 1024 * 1024 * 1024 * 1024):
        # Add psel points to the selection, with the rank specified as rank_psel.
        # If the point is already selected with lower rank, the rank is unchanged.
        c.add_psel_fsel(self, psel, rank_psel)
        self.update()

    def update(self, n_per_tslice = -1):
        # if n_per_tslice < 0: only update various indices
        # if n_per_tslice >= 0: only update parameters (n_per_tslice and prob)
        c.update_fsel(self, n_per_tslice)

    def select_rank_range(self, rank_start = 0, rank_stop = -1):
        # return new fsel with selected points that
        # rank_start <= rank and (rank < rank_stop or rank_stop == -1)
        # Does NOT change the n_per_tslice parameter for the new fsel
        fsel = FieldSelection()
        c.select_rank_range_fsel(fsel, self, rank_start, rank_stop)
        fsel.update()
        fsel.update(self.n_per_tslice())
        return fsel

    def select_t_range(self, rank_start = 0, rank_stop = -1):
        # return new fsel with selected points that
        # t_start <= t and (t < t_stop or t_stop == -1)
        # rank_start <= rank < rank_stop (rank_stop = -1 implies unlimited)
        # Does NOT change the n_per_tslice parameter for the new fsel
        fsel = FieldSelection()
        c.select_rank_range_fsel(fsel, self, rank_start, rank_stop)
        fsel.update()
        fsel.update(self.n_per_tslice())
        return fsel

    def to_psel(self):
        psel = PointSelection(None, self.geo())
        c.set_psel_fsel(psel, self)
        return psel

    def to_psel_local(self):
        psel = PointSelection(None, self.geo())
        c.set_psel_fsel_local(psel, self)
        return psel

    def save(self, path):
        return c.save_fsel(self, path)

    def load(self, path, n_per_tslice):
        return c.load_fsel(self, path, n_per_tslice)

    def geo(self):
        geo = Geometry((0, 0, 0, 0))
        c.set_geo_fsel(geo, self)
        return geo

    def total_site(self):
        return c.get_total_site_fsel(self)

    def n_elems(self):
        return c.get_n_elems_fsel(self)

    def n_per_tslice(self):
        return c.get_n_per_tslice_fsel(self)

    def prob(self):
        # return fsel.prob
        # n_per_tslice / spatial_volume
        return c.get_prob_fsel(self)

    def idx_from_coordinate(xg):
        return c.get_idx_from_coordinate_fsel(self, xg)

    def coordinate_from_idx(idx):
        return c.get_coordinate_from_idx_fsel(self, idx)

###

def is_matching_fsel(fsel1, fsel2):
    return c.is_matching_fsel(fsel1, fsel2)



from .geometry cimport *

cdef extern from "qlat/field-utils.h" namespace "qlat":

    cdef cppclass Field[T]:
        Field()
        void init()
        void init(const Geometry& geo)
        void init(const Geometry& geo, int multiplicity)
        void init(const Field[T]& field)
        const Field[T]& operator=(const Field[T]& field) except +



from qlat_utils.cp cimport *
from . cimport everything as cqlat

cdef class ElemType:

    pass

cdef class Geometry:

    cdef cqlat.Geometry xx

    cdef readonly long cdata

cdef class FieldBase:

    pass

cdef class SelectedField:

    pass

cdef class SelectedPoints:

    pass



from qlat_utils import *
import qlat.c as c
from qlat.c import get_size_node, get_coor_node
from qlat.c import sync_node

import numpy as np

def begin(*args):
    c.begin(*args)

def end():
    clean_cache()
    c.end()

@timer
def glb_sum_np(x):
    from qlat_utils.lat_io import LatData
    # x does NOT change
    shape = x.shape
    dtype = x.dtype
    l = list(x.flatten())
    ld = LatData()
    if dtype == np.dtype('float64'):
        ld.from_list(l, is_complex = False)
    elif dtype == np.dtype('int64'):
        ld.from_list(list(map(float, l)), is_complex = False)
    elif dtype == np.dtype('complex128'):
        ld.from_list(l, is_complex = True)
    else:
        displayln(dtype)
        assert False
    ld.glb_sum_in_place()
    return np.array(ld.to_list(), dtype = dtype).reshape(shape)

@timer
def glb_sum(x):
    # x does NOT change
    if isinstance(x, float):
        return c.glb_sum_double(x)
    elif isinstance(x, complex):
        return c.glb_sum_complex(x)
    elif isinstance(x, int):
        return c.glb_sum_long(x)
    elif isinstance(x, np.ndarray):
        return glb_sum_np(x)
    elif isinstance(x, list):
        return [ glb_sum(x_i) for x_i in x ]
    elif isinstance(x, tuple):
        return tuple([ glb_sum(x_i) for x_i in x ])
    else:
        # possible types: q.LatData
        return x.glb_sum()

@timer_verbose
def show_machine():
    displayln(f"id_node: {get_id_node():4} / {get_num_node()}"
            f" ; coor_node: {str(get_coor_node()):9}"
            f" / {str(get_size_node())}")



import ctypes
import sys
flags = sys.getdlopenflags()
sys.setdlopenflags(flags | ctypes.RTLD_GLOBAL)

from cqlat_utils import *
from qlat_utils.cp import *
from cqlat import *
from .cp import *

sys.setdlopenflags(flags)



from qlat_utils import *
import qlat.c as c

from qlat.geometry import *
from qlat.field import *
from qlat.field_selection import *
from qlat.utils_io import *

class SelectedField:

    # self.cdata
    # self.ctype
    # self.fsel

    def __init__(self, ctype, fsel, multiplicity = None):
        assert issubclass(ctype, c.ElemType)
        assert fsel is None or isinstance(fsel, FieldSelection)
        self.ctype = ctype
        self.fsel = fsel
        if fsel is None or multiplicity is None:
            self.cdata = c.mk_sfield(ctype)
        else:
            assert isinstance(multiplicity, int)
            self.cdata = c.mk_sfield_fsel(ctype, fsel, multiplicity)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_sfield(self)

    def __imatmul__(self, f1):
        # won't change self.fsel
        assert f1.ctype is self.ctype
        if isinstance(f1, SelectedField):
            # two fsel do not need to match
            if self.fsel is f1.fsel:
                c.set_sfield(self, f1)
            else:
                c.set_sfield_sfield(self, f1)
        elif isinstance(f1, FieldBase):
            c.set_sfield_field(self, f1)
        elif isinstance(f1, SelectedPoints):
            # psel may not have to be subset of fsel
            c.set_sfield_spfield(self, f1)
        else:
            assert False
        return self

    def copy(self, is_copying_data = True):
        f = SelectedField(self.ctype, self.fsel)
        if is_copying_data:
            f @= self
        return f

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def swap(self, x):
        assert isinstance(x, SelectedField)
        assert x.ctype is self.ctype
        assert x.fsel is self.fsel
        cdata = x.cdata
        x.cdata = self.cdata
        self.cdata = cdata

    def n_elems(self):
        return c.get_n_elems_sfield(self)

    def total_site(self):
        return c.get_total_site_sfield(self)

    def multiplicity(self):
        return c.get_multiplicity_sfield(self)

    def geo(self):
        geo = Geometry((0, 0, 0, 0))
        c.set_geo_sfield(geo, self)
        return geo

    def __iadd__(self, f1):
        assert isinstance(f1, SelectedField)
        assert f1.ctype is self.ctype
        c.set_add_sfield(self, f1)
        return self

    def __isub__(self, f1):
        assert isinstance(f1, SelectedField)
        assert f1.ctype is self.ctype
        c.set_sub_sfield(self, f1)
        return self

    def __imul__(self, factor):
        assert isinstance(factor, float)
        c.set_mul_double_sfield(self, factor)
        return self

    def set_zero(self):
        c.set_zero_sfield(self)

    def set_rand(self, rng, upper = 1.0, lower = 0.0):
        assert isinstance(rng, RngState)
        if self.ctype in field_ctypes_double:
            c.set_u_rand_double_sfield(self, rng, upper, lower)
        else:
            assert False

    def qnorm(self):
        return c.qnorm_sfield(self)

    def get_elems(self, idx):
        return np.array(c.get_elems_sfield(self, idx))

    def get_elem(self, idx, m = None):
        if m is None:
            return np.array(c.get_elem_sfield(self, idx))
        else:
            return np.array(c.get_elem_sfield(self, idx, m))

    def set_elems(self, idx, val):
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(val, bytes):
            return c.set_elems_sfield(self, idx, val)
        elif isinstance(val, np.ndarray):
            return self.set_elems(idx, val.tobytes())
        else:
            assert False

    def set_elem(self, idx, m, val):
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(val, bytes):
            return c.set_elem_sfield(self, idx, m, val)
        elif isinstance(val, np.ndarray):
            return self.set_elem(idx, m, val.tobytes())
        else:
            assert False

    def __getitem__(self, i):
        # i can be (idx, m,) or idx
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], int):
            idx, m = i
            return self.get_elem(idx, m)
        elif isinstance(i, int):
            idx = i
            return self.get_elems(idx)
        else:
            assert False
            return None

    def __setitem__(self, i, val):
        # i can be (idx, m,) or idx
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], int):
            idx, m = i
            return self.set_elem(idx, m, val)
        elif isinstance(i, int):
            idx = i
            return self.set_elems(idx, val)
        else:
            assert False
            return None

    def sparse(self, sel):
        # deprecated
        displayln_info("SelectedField.sparse: deprecated")
        if isinstance(sel, PointSelection):
            from qlat.selected_points import SelectedPoints, set_selected_points
            psel = sel
            sp = SelectedPoints(self.ctype, psel)
            set_selected_points(sp, self)
            return sp
        elif isinstance(sel, FieldSelection):
            fsel = sel
            sf = SelectedField(self.ctype, fsel)
            set_selected_field(sf, self)
            return sf
        else:
            raise Exception("Field.sparse")

    def save_direct(self, path, *args):
        # possible way to call:
        # f.save_direct(path)
        # f.save_direct(path, new_size_node)
        # f.save_direct(sfw, fn)
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            assert len(args) == 0
            n_bytes = c.save_sfield(self, path)
        elif isinstance(path, ShuffledFieldsWriter):
            sfw = path
            [fn] = args
            n_bytes = sfw.write(fn, self)
        else:
            raise Exception("SelectedField.save_direct")
        assert n_bytes != 0
        return n_bytes

    def load_direct(self, path, *args):
        # possible way to call:
        # f.load_direct(path)
        # f.load_direct(sfr, fn)
        # if self.fsel is None, self.fsel will be set during f.load_direct(sfr, fn)
        from qlat.fields_io import ShuffledFieldsReader
        if isinstance(path, str):
            n_bytes = c.load_sfield(self, path)
        elif isinstance(path, ShuffledFieldsReader):
            sfr = path
            [fn] = args
            n_bytes = sfr.read(fn, self)
        else:
            raise Exception("SelectedField.load")
        assert n_bytes != 0
        return n_bytes

    def save_64(self, path, *args):
        f = self.copy()
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            f.to_from_endianness("big_64")
        elif isinstance(path, ShuffledFieldsWriter):
            f.to_from_endianness("little_64")
        return f.save_direct(path, *args)

    def save_double(self, path, *args):
        return self.save_64(path, *args)

    def save_float_from_double(self, path, *args):
        ff = SelectedField(c.ElemTypeFloat, self.fsel)
        ff.float_from_double(self)
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            ff.to_from_endianness("big_32")
        elif isinstance(path, ShuffledFieldsWriter):
            ff.to_from_endianness("little_32")
        return ff.save_direct(path, *args)

    def load_64(self, path, *args):
        ret = self.load_direct(path, *args)
        if ret > 0:
            from qlat.fields_io import ShuffledFieldsReader
            if isinstance(path, str):
                self.to_from_endianness("big_64")
            elif isinstance(path, ShuffledFieldsReader):
                self.to_from_endianness("little_64")
        return ret

    def load_double(self, path, *args):
        return self.load_64(path, *args)

    def load_double_from_float(self, path, *args):
        ff = SelectedField(c.ElemTypeFloat, self.fsel)
        ret = ff.load_direct(path, *args)
        if ret > 0:
            from qlat.fields_io import ShuffledFieldsReader
            if isinstance(path, str):
                ff.to_from_endianness("big_32")
            elif isinstance(path, ShuffledFieldsReader):
                ff.to_from_endianness("little_32")
            self.double_from_float(ff)
        return ret

    def float_from_double(self, f):
        assert isinstance(f, SelectedField)
        assert self.ctype == c.ElemTypeFloat
        self.fsel = f.fsel
        c.convert_float_from_double_sfield(self, f)

    def double_from_float(self, ff):
        assert isinstance(ff, SelectedField)
        assert ff.ctype == c.ElemTypeFloat
        self.fsel = ff.fsel
        c.convert_double_from_float_sfield(self, ff)

    def to_from_endianness(self, tag):
        assert isinstance(tag, str)
        c.to_from_endianness_sfield(self, tag)

    def field_shift(self, shift, is_reflect = False):
        # return new shifted SelectedField
        # shift is the coordinate to shift the field
        # is_reflect determine whether to negate coordinate after shift
        f1 = self.copy(is_copying_data = False)
        f1.fsel = FieldSelection()
        c.field_shift_sfield(f1, self, shift, is_reflect)
        return f1

    def glb_sum_tslice(self, *, t_dir = 3):
        # return SelectedPoints(self.ctype, get_psel_tslice(self.total_site(), t_dir = t_dir))
        from qlat.selected_points import SelectedPoints
        psel = get_psel_tslice(self.total_site(), t_dir = t_dir)
        sp = SelectedPoints(self.ctype, psel)
        if self.ctype in field_ctypes_double:
            c.glb_sum_tslice_double_sfield(sp, self, t_dir)
        elif self.ctype in field_ctypes_long:
            c.glb_sum_tslice_long_sfield(sp, self, t_dir)
        else:
            assert False
        return sp

###

@timer
def set_selected_field(sf, f):
    # deprecated use @=
    displayln_info("set_selected_field: deprecated")
    assert isinstance(sf, SelectedField)
    if isinstance(f, FieldBase):
        c.set_sfield_field(sf, f)
    elif isinstance(f, SelectedField):
        c.set_sfield_sfield(sf, f)
    else:
        raise Exception("set_selected_field")



cdef class ElemTypeTYPENAME(ElemType):

    name = "TYPENAME"

### -------------------------------------------------------------------

cdef class FieldTYPENAME(FieldBase):

    ctype = ElemTypeTYPENAME

    def __cinit__(self):
        self.cdata = <long>&(self.xx)

    def __init__(self, geo = None, multiplicity = None):
        if geo is None:
            return
        assert isinstance(geo, Geometry)
        if multiplicity is None:
            self.xx.init((<Geometry>geo).xx)
        else:
            assert isinstance(multiplicity, int)
            self.xx.init((<Geometry>geo).xx, <int>multiplicity)

### -------------------------------------------------------------------

field_type_dict[ElemTypeTYPENAME] = FieldTYPENAME

### -------------------------------------------------------------------



#    Qlattice (https://github.com/waterret/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import copy
import qlat as q

class AmaVal:

    def __init__(self, val = None, corrections = None):
        # should have the following:
        # self.val: sloppy value or None
        # self.corrections: list [ (val, description_dict,), ... ]
        # description_dict:
        # { source_specification:
        #   (accuracy_level_relative_to_the_basic_accuracy,
        #    probability_of_having_this_accuracy,),
        # }
        # self.corrections should include the sloppy value and self.val is optional.
        self.val = val
        if corrections is None:
            self.corrections = []
        else:
            self.corrections = corrections

    def __repr__(self):
        return f"AmaVal({self.val},{self.corrections})"

    def __mul__(self, other):
        return ama_msc_mult(self, other)

    def __rmul__(self, other):
        return ama_msc_mult(other, self)

    def __add__(self, other):
        return ama_msc_add(self, other)

    def __radd__(self, other):
        return ama_msc_add(other, self)

###

@q.timer
def mk_ama_val(val, source_specification, val_list, rel_acc_list, prob_list):
    # source_specification need to be unique for each propagator source to ensure proper AMA correction for final result
    # e.g. source_specification = ("point", (12, 2, 3, 4,),)
    # val is the sloppy result
    # val == val_list[0]
    # None value in val_list will be removed automatically
    assert len(val_list) == len(rel_acc_list)
    assert len(val_list) == len(prob_list)
    if val is not val_list[0]:
        assert val == val_list[0]
    corrections = []
    for val_i, rel_acc_i, prob_i in zip(val_list, rel_acc_list, prob_list):
        if val_i is not None:
            corrections.append((val_i, { source_specification: (rel_acc_i, prob_i), },))
    if len(corrections) == 1:
        return val
    return AmaVal(val, corrections)

@q.timer
def ama_apply1_corrections(f, x):
    assert isinstance(x, AmaVal)
    corrections = [ (f(v), d,) for v, d in x.corrections ]
    return corrections

def ama_apply1(f, x):
    if not isinstance(x, AmaVal):
        return f(x)
    else:
        return AmaVal(None, ama_apply1_corrections(f, x))

def ama_counts(x):
    # counts how many times need to compute the val
    if not isinstance(x, AmaVal):
        return 1
    else:
        return len(x.corrections)

def merge_description_dict(d1, d2):
    sd1 = set(d1)
    sd2 = set(d2)
    common_keys = sd1 & sd2
    for key in common_keys:
        if d1[key] != d2[key]:
            return None
    new_keys = sd2 - sd1
    d = d1.copy()
    for key in new_keys:
        d[key] = d2[key]
    return d

@q.timer
def ama_apply2_ama_val(f, x, y):
    assert isinstance(x, AmaVal)
    assert isinstance(y, AmaVal)
    corrections = []
    for v_x, d_x in x.corrections:
        for v_y, d_y in y.corrections:
            d = merge_description_dict(d_x, d_y)
            if d is not None:
                corrections.append((f(v_x, v_y), d,))
    return AmaVal(None, corrections)

@q.timer
def ama_apply2_r_ama_val(f, x, y):
    assert isinstance(y, AmaVal)
    corrections = [ (f(x, v), d,) for v, d in y.corrections ]
    return AmaVal(None, corrections)

@q.timer
def ama_apply2_l_ama_val(f, x, y):
    assert isinstance(x, AmaVal)
    corrections = [ (f(v, y), d,) for v, d in x.corrections ]
    return AmaVal(None, corrections)

def ama_apply2_r(f, x, y):
    if not isinstance(y, AmaVal):
        return f(x, y)
    else:
        return ama_apply2_r_ama_val(f, x, y)

def ama_apply2_l(f, x, y):
    if not isinstance(x, AmaVal):
        return f(x, y)
    else:
        return ama_apply2_l_ama_val(f, x, y)

def ama_apply2(f, x, y):
    if not isinstance(x, AmaVal) and not isinstance(y, AmaVal):
        return f(x, y)
    elif isinstance(x, AmaVal) and not isinstance(y, AmaVal):
        return ama_apply2_l_ama_val(f, x, y)
    elif not isinstance(x, AmaVal) and isinstance(y, AmaVal):
        return ama_apply2_r_ama_val(f, x, y)
    else:
        return ama_apply2_ama_val(f, x, y)

@q.timer
def ama_list(*args):
    l = len(args)
    assert l >= 0
    def f_add(rs, x):
        return rs + [ x, ]
    res = []
    for x in args:
        res = ama_apply2(f_add, res, x)
    return res

def ama_apply(f, *args):
    res = ama_list(*args)
    def f_list(xs):
        return f(*xs)
    return ama_apply1(f_list, res)

@q.timer
def ama_extract_ama_val(x, *, is_sloppy = False):
    corrections = x.corrections
    assert isinstance(corrections, list)
    assert corrections
    # keys = [ source_specification, ... ]
    keys = list(corrections[0][1].keys())
    def get_level_prob(key):
        s = set([ d[key] for v, d in corrections ])
        return sorted(list(s))
    # dict_level_prob[key] = list of accuracy level and probability_of_having_this_accuracy pairs for this key (source_specification)
    dict_level_prob = { k: get_level_prob(k) for k in keys }
    for k, v in dict_level_prob.items():
        assert len(v) >= 2
    # dict_level[key] = list of accuracy levels for this key (source_specification)
    dict_level = { k: [ l for l, prob in v ] for k, v in dict_level_prob.items() }
    # dict_prob[key] = list of probability_of_having_this_accuracy
    dict_prob = { k: [ prob for l, prob in v ] for k, v in dict_level_prob.items() }
    # dict_val[(level, ...)] = v
    dict_val = { tuple([ d[k][0] for k in keys ]): v for v, d in corrections }
    # initial val is sloppy only
    if is_sloppy or (not keys):
        return copy.copy(dict_val[tuple([ dict_level[key][0] for key in keys ])])
    def ama_corr(fixed_levels, remaining_keys):
        if not remaining_keys:
            return dict_val[tuple(fixed_levels)]
        else:
            key = remaining_keys[0]
            rest_keys = remaining_keys[1:]
            levels = dict_level[key]
            probs = dict_prob[key]
            vals = [ ama_corr(fixed_levels + [ l, ], rest_keys) for l in levels ]
            val = copy.copy(vals[0])
            for i in range(1, len(vals)):
                diff = copy.copy(vals[i])
                diff -= vals[i - 1]
                diff *= 1 / probs[i]
                val += diff
            return val
    return ama_corr([], keys)

def ama_extract(x, *, is_sloppy = False):
    if not isinstance(x, AmaVal):
        return x
    elif isinstance(x, AmaVal):
        return ama_extract_ama_val(x, is_sloppy = is_sloppy)
    else:
        assert False

def ama_msc_mult(x, y):
    def f(x, y):
        return x * y
    return ama_apply2(f, x, y)

def ama_msc_add(x, y):
    def f(x, y):
        return x + y
    return ama_apply2(f, x, y)

if __name__ == "__main__":
    v1 = mk_ama_val(1.0, "x", [ 1.0, 1.01, 1.011, ], [ 0, 1, 2, ], [ 1.0, 0.1, 0.02, ])
    v2 = mk_ama_val(2.0, "y", [ 2.0, 2.01, 2.011, ], [ 0, 1, 2, ], [ 1.0, 0.1, 0.02, ])
    v3 = mk_ama_val(3.0, "z", [ 3.0, None, None, ], [ 0, 1, 2, ], [ 1.0, 0.1, 0.02, ])
    def f(x, y):
        return x * y
    v4 = ama_apply2(f, v1, v2)
    v5 = ama_apply2(f, v1, v3)
    v6 = ama_apply2(f, v2, v2)
    print(ama_extract(v1), v1)
    print(ama_extract(v2), v2)
    print(ama_extract(v3), v3)
    print(ama_extract(v4), v4)
    print(ama_extract(v5), v5)
    print(ama_extract(v6), v6)



#    Qlattice (https://github.com/waterret/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from auto_contractor.compile import *
from auto_contractor.ama import *

from qlat_utils import \
        rel_mod, rel_mod_sym, c_rel_mod_sqr

from qlat_utils.cp import \
        as_wilson_matrix, as_wilson_matrix_g5_herm

import numpy as np
import qlat as q
import copy
import cmath
import math
import importlib
import time
import os
import subprocess

def load_prop(x):
    if isinstance(x, tuple):
        assert len(x) == 2 and x[0] == "g5_herm"
        return ama_apply1(as_wilson_matrix_g5_herm, x[1])
    return ama_apply1(as_wilson_matrix, x)

@q.timer
def get_cexpr_names(cexpr):
    names = [ name for name, expr in cexpr.named_exprs ]
    return names

def eval_cexpr_get_props(cexpr : CExpr, *, positions_dict, get_prop):
    assert cexpr.function is not None
    return cexpr.function["cexpr_function_get_prop"](positions_dict, get_prop)

def eval_cexpr_eval(cexpr : CExpr, *, positions_dict, props):
    assert cexpr.function is not None
    return cexpr.function["cexpr_function_eval"](positions_dict, props)

def eval_cexpr(cexpr : CExpr, *, positions_dict, get_prop, is_ama_and_sloppy = False):
    # return 1 dimensional np.array
    # cexpr can be cexpr object or can be a compiled function
    # xg = positions_dict[position]
    # mat_mspincolor = get_prop(flavor, xg_snk, xg_src)
    # e.g. ("point-snk", [ 1, 2, 3, 4, ]) = positions_dict["x_1"]
    # e.g. flavor = "l"
    # e.g. xg_snk = ("point-snk", [ 1, 2, 3, 4, ])
    # if is_ama_and_sloppy: return (val_ama, val_sloppy,)
    # if not is_ama_and_sloppy: return val_ama
    # interface function
    assert cexpr.function is not None
    return cexpr.function["cexpr_function"](positions_dict = positions_dict, get_prop = get_prop, is_ama_and_sloppy = is_ama_and_sloppy)

meson_build_content = """\
project('qlat-auto-contractor-cexpr', 'cpp', 'cython',
  version: '1.0',
  license: 'GPL-3.0-or-later',
  default_options: [
    'warning_level=3',
    'cpp_std=c++14',
    'libdir=lib',
    'optimization=2',
    'debug=false',
    'cython_language=cpp',
    ])
add_project_arguments('-fno-strict-aliasing', language: ['c', 'cpp'])
cpp = meson.get_compiler('cpp')
py_mod = import('python')
py3 = py_mod.find_installation('python3')
message(py3.path())
message(py3.get_install_dir())
omp = dependency('openmp').as_system()
zlib = dependency('zlib').as_system()
fftw = dependency('fftw3').as_system()
fftwf = dependency('fftw3f').as_system()
message('fftw libdir', fftw.get_variable('libdir'))
message('fftwf libdir', fftwf.get_variable('libdir'))
fftw_all = [ fftw, fftwf, ]
math = cpp.find_library('m')
numpy_include = run_command(py3.path(), '-c', 'import numpy as np ; print(np.get_include())', check: true).stdout().strip()
message('numpy include', numpy_include)
numpy = declare_dependency(
  include_directories:  include_directories(numpy_include),
  dependencies: [ py3.dependency(), ],
  ).as_system()
qlat_utils_include = run_command(py3.path(), '-c', 'import qlat_utils as q ; print("\\\\n".join(q.get_include_list()))', env: environment({'q_verbose': '-1'}), check: true).stdout().strip().split('\\n')
message('qlat_utils include', qlat_utils_include)
qlat_utils_lib = run_command(py3.path(), '-c', 'import qlat_utils as q ; print("\\\\n".join(q.get_lib_list()))', env: environment({'q_verbose': '-1'}), check: true).stdout().strip().split('\\n')
message('qlat_utils lib', qlat_utils_lib)
qlat_utils = declare_dependency(
  include_directories:  include_directories(qlat_utils_include),
  dependencies: [
    py3.dependency().as_system(),
    cpp.find_library('qlat-utils', dirs: qlat_utils_lib),
    numpy, omp, zlib, math, ],
  )
deps = [ qlat_utils, fftw_all, ]
if not cpp.check_header('Eigen/Eigen')
  eigen = dependency('eigen3').as_system()
  deps += [ eigen, ]
endif
incdir = []
codelib = py3.extension_module('code',
  files('code.pyx'),
  dependencies: deps,
  include_directories: incdir,
  install: false,
  )
"""

@q.timer
def cache_compiled_cexpr(calc_cexpr, path):
    # Obtain: cexpr = calc_cexpr().
    # Save cexpr object in pickle format for future reuse.
    # Generate python code and save for future reuse
    # Load the python module and assign function and total_sloppy_flops
    # Return fully loaded cexpr
    # interface function
    fn_pickle = path + "/data.pickle"
    fn_code = path + "/build/code.so"
    @q.timer
    def compile_cexpr_meson_setup():
        subprocess.run(["meson", "setup", "build"], cwd = path)
    @q.timer
    def compile_cexpr_meson_compile():
        subprocess.run(["meson", "compile", "-C", "build"], cwd = path)
    @q.timer
    def calc_compile_cexpr():
        cexpr_original = calc_cexpr()
        cexpr_optimized = copy.deepcopy(cexpr_original)
        cexpr_optimized.optimize()
        code_py = cexpr_code_gen_py(cexpr_optimized)
        fn_py = path + "/code.pyx"
        q.qtouch_info(fn_py, code_py)
        content_original = display_cexpr(cexpr_original)
        q.qtouch_info(path + "/cexpr.original.txt", content_original)
        content_optimized = display_cexpr(cexpr_optimized)
        q.qtouch_info(path + "/cexpr.optimized.txt", content_optimized)
        q.qtouch_info(path + "/meson.build", meson_build_content)
        compile_cexpr_meson_setup()
        compile_cexpr_meson_compile()
        data = dict()
        data["cexpr_optimized"] = cexpr_optimized
        data["cexpr_original"] = cexpr_original
        data["code_py"] = code_py
        q.save_pickle_obj(data, fn_pickle)
        return cexpr_optimized
    if q.get_id_node() == 0 and not q.does_file_exist(fn_pickle):
        calc_compile_cexpr()
    q.sync_node()
    while not q.does_file_exist(fn_pickle):
        q.displayln(3, f"cache_compiled_cexpr: Node {q.get_id_node()}: waiting for '{fn_pickle}'.")
        time.sleep(0.1)
    cexpr = q.load_pickle_obj(fn_pickle)["cexpr_optimized"]
    module = importlib.import_module((path + "/build/code").replace("/", "."))
    cexpr.function = {
            # cexpr_function(positions_dict, get_prop, is_ama_and_sloppy = False) => val as 1-D np.array
            "cexpr_function" : module.cexpr_function,
            # cexpr_function_get_prop(positions_dict, get_prop) => props
            "cexpr_function_get_prop" : module.cexpr_function_get_prop,
            # cexpr_function_eval(positions_dict, props) => ama_val as AmaVal of 1-D np.array
            "cexpr_function_eval" : module.cexpr_function_eval,
            }
    cexpr.total_sloppy_flops = module.total_sloppy_flops
    return cexpr

def make_rand_spin_color_matrix(rng_state):
    rs = rng_state
    wm = q.WilsonMatrix()
    wm_arr = np.asarray(wm)
    wm_arr[:] = np.array(
            [ rs.u_rand_gen() + 1j * rs.u_rand_gen() for i in range(144) ],
            dtype = complex).reshape(12, 12)
    return wm

def make_rand_spin_matrix(rng_state):
    rs = rng_state
    sm = q.SpinMatrix()
    sm_arr = np.asarray(sm)
    sm_arr[:] = np.array(
            [ rs.u_rand_gen() + 1j * rs.u_rand_gen() for i in range(16) ],
            dtype = complex).reshape(4, 4)
    return sm

def benchmark_show_check(check):
    return " ".join([ f"{v:.10E}" for v in check ])

@q.timer
def benchmark_eval_cexpr(cexpr : CExpr, *,
        benchmark_size = 10,
        benchmark_num = 10,
        benchmark_num_ama = 2,
        benchmark_rng_state = None):
    if benchmark_rng_state is None:
        benchmark_rng_state = q.RngState("benchmark_eval_cexpr")
    expr_names = get_cexpr_names(cexpr)
    n_expr = len(expr_names)
    n_pos = len(cexpr.positions)
    prop_dict = {}
    size = q.Coordinate([ 8, 8, 8, 16, ])
    positions = [
            ("point", tuple(benchmark_rng_state.split(f"positions {pos_idx}").c_rand_gen(size).list()),)
            for pos_idx in range(n_pos)
            ]
    set_flavors = set()
    for name, value in cexpr.variables_prop:
        assert name.startswith("V_S_")
        x = value
        assert isinstance(x, Op)
        assert x.otype == "S"
        flavor = x.f
        set_flavors.add(flavor)
    for flavor in set_flavors:
        for pos_src_idx in range(n_pos):
            pos_src = positions[pos_src_idx]
            for pos_snk_idx in range(n_pos):
                pos_snk = positions[pos_snk_idx]
                prop = make_rand_spin_color_matrix(benchmark_rng_state.split(f"prop {flavor} {pos_snk_idx} {pos_src_idx}"))
                prop_ama = make_rand_spin_color_matrix(benchmark_rng_state.split(f"prop ama {flavor} {pos_snk_idx} {pos_src_idx}"))
                prop_dict[(flavor, pos_snk, pos_src,)] = mk_ama_val(prop, pos_src, [ prop, prop_ama, ], [ 0, 1, ], [ 1.0, 0.5, ])
    def mk_pos_dict(k):
        positions_dict = {}
        positions_dict["size"] = size.list()
        idx_list = q.random_permute(list(range(n_pos)), benchmark_rng_state.split(f"pos_dict {k}"))
        for pos, idx in zip(cexpr.positions, idx_list):
            positions_dict[pos] = positions[idx]
        return positions_dict
    positions_dict_list = [ mk_pos_dict(k) for k in range(benchmark_size) ]
    #
    @q.timer
    def get_prop(flavor, pos_snk, pos_src):
        return ama_extract(prop_dict[(flavor, pos_snk, pos_src)], is_sloppy = True)
    @q.timer_verbose
    def benchmark_eval_cexpr_run():
        res_list = []
        for k in range(benchmark_size):
            res = eval_cexpr(cexpr, positions_dict = positions_dict_list[k], get_prop = get_prop)
            res_list.append(res)
        res = np.array(res_list)
        assert res.shape == (benchmark_size, n_expr,)
        return res
    @q.timer
    def get_prop_ama(flavor, pos_snk, pos_src):
        return prop_dict[(flavor, pos_snk, pos_src)]
    @q.timer_verbose
    def benchmark_eval_cexpr_run_with_ama():
        res_list = []
        for k in range(benchmark_size):
            res1 = eval_cexpr(cexpr, positions_dict = positions_dict_list[k], get_prop = get_prop_ama)
            res2 = eval_cexpr(cexpr, positions_dict = positions_dict_list[k], get_prop = get_prop)
            res_ama, res_sloppy = eval_cexpr(cexpr, positions_dict = positions_dict_list[k], get_prop = get_prop_ama, is_ama_and_sloppy = True)
            assert q.qnorm(res1 - res_ama) == 0
            assert q.qnorm(res2 - res_sloppy) == 0
            res_list.append(res_ama)
        res = np.array(res_list)
        assert res.shape == (benchmark_size, n_expr,)
        return res
    def mk_check_vector(k):
        rs = benchmark_rng_state.split(f"check_vector {k}")
        res = np.array([
            [ complex(rs.u_rand_gen(1.0, -1.0), rs.u_rand_gen(1.0, -1.0)) for i in range(n_expr) ]
            for k in range(benchmark_size) ])
        return res
    check_vector_list = [ mk_check_vector(k) for k in range(3) ]
    def check_res(res):
        return [ np.tensordot(res, cv).item() for cv in check_vector_list ]
    q.displayln_info(f"benchmark_eval_cexpr: benchmark_size={benchmark_size}")
    q.timer_fork(0)
    check = None
    for i in range(benchmark_num):
        res = benchmark_eval_cexpr_run()
        new_check = check_res(res)
        if check is None:
            check = new_check
        else:
            assert check == new_check
    check_ama = None
    for i in range(benchmark_num_ama):
        res_ama = benchmark_eval_cexpr_run_with_ama()
        new_check_ama = check_res(res_ama)
        if check_ama is None:
            check_ama = new_check_ama
        else:
            assert check_ama == new_check_ama
    q.timer_display()
    q.timer_merge()
    q.displayln_info(f"benchmark_eval_cexpr: {benchmark_show_check(check)} {benchmark_show_check(check_ama)}")
    return check, check_ama

def sqr_component(x):
    return x.real * x.real + 1j * x.imag * x.imag

def sqrt_component(x):
    return math.sqrt(x.real) + 1j * math.sqrt(x.imag)

def sqr_component_array(arr):
    return np.array([ sqr_component(x) for x in arr ])

def sqrt_component_array(arr):
    return np.array([ sqrt_component(x) for x in arr ])

def get_mpi_chunk(total_list, *, rng_state = None):
    # rng_state has to be the same on all the nodes
    # e.g. rng_state = q.RngState("get_mpi_chunk")
    total = len(total_list)
    id_worker = q.get_id_node()
    num_worker = q.get_num_node()
    size_max = (total - 1) // num_worker + 1;
    start = min(id_worker * size_max, total);
    stop = min(start + size_max, total);
    # size = stop - start;
    if rng_state is not None:
        assert isinstance(rng_state, q.RngState)
        total_list = q.random_permute(total_list, rng_state)
    return total_list[start:stop]
